{"status_code": 200, "headers": {"content-type": "text/plain; charset=utf-8", "content-length": "2087", "connection": "keep-alive", "date": "Wed, 14 Aug 2024 07:50:49 GMT", "x-powered-by": "huggingface-moon", "cross-origin-opener-policy": "same-origin", "referrer-policy": "strict-origin-when-cross-origin", "x-request-id": "Root=1-66bc61d9-1ce959d003aeb98265c16506;0be4b2ee-eb49-4721-afa9-8d6d1d262d22", "access-control-allow-origin": "https://huggingface.co", "vary": "Origin", "access-control-expose-headers": "X-Repo-Commit,X-Request-Id,X-Error-Code,X-Error-Message,X-Total-Count,ETag,Link,Accept-Ranges,Content-Range", "x-repo-commit": "0547ed36a86561e2e42fecec8fd0c4f6953e33c4", "accept-ranges": "bytes", "content-disposition": "inline; filename*=UTF-8''configuration_qwen.py; filename=\"configuration_qwen.py\";", "content-security-policy": "default-src none; sandbox", "etag": "\"2ccfc926ca2615796fd0822d0bc2e140a38aa1f7\"", "x-cache": "Miss from cloudfront", "via": "1.1 2b4afc9aa87e1c567d185069953d85bc.cloudfront.net (CloudFront)", "x-amz-cf-pop": "HKG1-P1", "x-amz-cf-id": "-G96rUPlO9eOetqetzsokEIuXR8O1QxDxIkiE9yY1-oDPgBTVR0QDA=="}, "content": ""}