{"status_code": 200, "headers": {"content-type": "text/plain; charset=utf-8", "content-length": "44654", "connection": "keep-alive", "date": "Wed, 14 Aug 2024 07:50:49 GMT", "x-powered-by": "huggingface-moon", "cross-origin-opener-policy": "same-origin", "referrer-policy": "strict-origin-when-cross-origin", "x-request-id": "Root=1-66bc61d9-6802815e19df7fd15f522c2e;8a3fbc62-fc2f-40ef-83c2-9cd73529857d", "access-control-allow-origin": "https://huggingface.co", "vary": "Origin", "access-control-expose-headers": "X-Repo-Commit,X-Request-Id,X-Error-Code,X-Error-Message,X-Total-Count,ETag,Link,Accept-Ranges,Content-Range", "x-repo-commit": "0547ed36a86561e2e42fecec8fd0c4f6953e33c4", "accept-ranges": "bytes", "content-disposition": "inline; filename*=UTF-8''modeling_qwen.py; filename=\"modeling_qwen.py\";", "content-security-policy": "default-src none; sandbox", "etag": "\"67eee3163115ab2a5b824b8a571dc9b551d878e8\"", "x-cache": "Miss from cloudfront", "via": "1.1 ca87bf429f214c70d8beaa0c7de1e702.cloudfront.net (CloudFront)", "x-amz-cf-pop": "HKG1-P1", "x-amz-cf-id": "FxQ-_f3yPqp7_eEWiDGrHhyWKgZmZe_lWDJ64SDpS-RH7jHARx6Cig=="}, "content": ""}