{"status_code": 200, "headers": {"content-type": "text/plain; charset=utf-8", "content-length": "14873", "connection": "keep-alive", "date": "Wed, 14 Aug 2024 07:52:40 GMT", "x-powered-by": "huggingface-moon", "cross-origin-opener-policy": "same-origin", "referrer-policy": "strict-origin-when-cross-origin", "x-request-id": "Root=1-66bc6248-7eb4420f3abfb2764b5a49cd;99ce04f9-703f-4f2a-872f-aa07cd7b9c34", "access-control-allow-origin": "https://huggingface.co", "vary": "Origin", "access-control-expose-headers": "X-Repo-Commit,X-Request-Id,X-Error-Code,X-Error-Message,X-Total-Count,ETag,Link,Accept-Ranges,Content-Range", "x-repo-commit": "0547ed36a86561e2e42fecec8fd0c4f6953e33c4", "accept-ranges": "bytes", "content-disposition": "inline; filename*=UTF-8''qwen_generation_utils.py; filename=\"qwen_generation_utils.py\";", "content-security-policy": "default-src none; sandbox", "etag": "\"911c6bd6967774bbd6150cab86b02ff101c5ad51\"", "x-cache": "Miss from cloudfront", "via": "1.1 c8300eed0230a129b2b5696d1d6037f0.cloudfront.net (CloudFront)", "x-amz-cf-pop": "HKG1-P1", "x-amz-cf-id": "wHlHhxnx_DGO_zjwghq0V5hBcE3cIrvMvWavQIEzUy1mHrl_tMKxOg=="}, "content": ""}