{"status_code": 200, "headers": {"content-type": "text/plain; charset=utf-8", "content-length": "14553", "connection": "keep-alive", "date": "Wed, 14 Aug 2024 07:52:47 GMT", "x-powered-by": "huggingface-moon", "cross-origin-opener-policy": "same-origin", "referrer-policy": "strict-origin-when-cross-origin", "x-request-id": "Root=1-66bc624f-42f4f9627e429c8631f6e08d;86091ebd-7959-4275-a6c7-fe266aaa1f65", "access-control-allow-origin": "https://huggingface.co", "vary": "Origin", "access-control-expose-headers": "X-Repo-Commit,X-Request-Id,X-Error-Code,X-Error-Message,X-Total-Count,ETag,Link,Accept-Ranges,Content-Range", "x-repo-commit": "0547ed36a86561e2e42fecec8fd0c4f6953e33c4", "accept-ranges": "bytes", "content-disposition": "inline; filename*=UTF-8''visual.py; filename=\"visual.py\";", "content-security-policy": "default-src none; sandbox", "etag": "\"85f3f7dcc3bc392b997bd1ee812f317334e9db55\"", "x-cache": "Miss from cloudfront", "via": "1.1 2f3726161bcad3fa1465262c0d76e9ac.cloudfront.net (CloudFront)", "x-amz-cf-pop": "HKG1-P1", "x-amz-cf-id": "xzfYu6cBqDDgwhrNTGctED4XaYewop2M4W9nhUWgfWM_ycFho7GUJA=="}, "content": ""}